// Created by:  Eric Sheesley (Inferno)
// Date:  09/06/01
// Email:  esheesle@bcpl.net
// This program allows netcat/cryptcat to call a remote program and actually interface with its stdin/stdout
// Hobbit's original simply spawned a shell and let it run but to be able to encrypt the contents we needed
// to interface with the data before it was sent or recieved.  TTY's are used to accomplish this.
// This should work on all *nix systems but hard to say for sure as I don't have all of them.  I have tested
// this on several flavors of linux and solaris and all works well.  Email me if you run into troubles with any
// other nix platforms.  NOTE:  The windows version already does something similar but in a completely
// different way.
// Thanks to the comp.unix.programmer group who helped with some tips when i had bugs to squash
// Features to add:
// -A (-L) switch like the windows version to allow for respawning of the session without needing to mess with
//   inetd
// -Allowing for multiple clients to connect to one server instance on the same port.

#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <termios.h>
#ifndef TIOCGWINSZ
#include <sys/ioctl.h>
#endif
#include <signal.h>
#include "doexec.h"

#define BUFFSIZE 512

static void sig_term(int);
static volatile sig_atomic_t sigcaught;
static void set_noecho(int);
void do_driver(char *);
void loop(int, int);

ssize_t writen(int fd, const void *vptr, size_t n)
{
	size_t nleft;
	size_t nwritten;
	const char *ptr;

	ptr = vptr;
	nleft = n;
	while (nleft > 0)
	{
		if ((nwritten = write(fd, ptr, nleft)) <= 0)
			return(nwritten);
		nleft -= nwritten;
		ptr += nwritten;
	}
	return(n);
}

Sigfunc *signal_intr(int signo, Sigfunc *func)
{
	struct sigaction act, oact;

	act.sa_handler = func;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;
#ifdef SA_INTERRUPT     // SunOS
   act.sa_flags |= SA_INTERRUPT;
#endif
	if (sigaction(signo, &act, &oact) < 0)
	   return(SIG_ERR);
   return(oact.sa_handler);
}

int ptym_open(char *pts_name)
{
	int fdm;
	char *ptr1, *ptr2;

	strcpy(pts_name, "/dev/ptyXY");
	for (ptr1 = "pqrstuvwxyzPQRST"; *ptr1 != 0; ptr1++)
	{
		pts_name[8] = *ptr1;
		for (ptr2 = "0123456789abcdef"; *ptr2 != 0; ptr2++)
		{
			pts_name[9] = *ptr2;
			// Try to open master tty
			if ((fdm = open(pts_name, O_RDWR)) < 0)
			{
				if (errno == ENOENT)
					return(-1);
				else
					continue;
			}
			pts_name[5] = 't';  // Change pty to tty
			return(fdm);
		}
	}
	return(-1);
}

int ptys_open(int fdm, char *pts_name)
{
	struct group *grptr;
	int gid, fds;

	if ((grptr = getgrnam("tty")) != NULL)
		gid = grptr->gr_gid;
	else
		gid = -1;
	// Must be root for the following
	chown(pts_name, getuid(), gid);
	chmod(pts_name, S_IRUSR | S_IWUSR | S_IWGRP);

	if ((fds = open(pts_name, O_RDWR)) < 0)
	{
		close(fdm);
		return(-1);
	}
	return(fds);
}

pid_t pty_fork(int *ptrfdm, char *slave_name, const struct termios *slave_termios, const struct winsize *slave_winsize)
{
	int fdm, fds;
	pid_t pid;
	char pts_name[20];

	if ((fdm = ptym_open(pts_name)) < 0)
		printf("can't open master pty: %s", pts_name);

	if (slave_name != NULL)
		strcpy(slave_name, pts_name);

	if ((pid = fork()) < 0)
		return(-1);
	else if (pid == 0)
	{
		if (setsid() < 0)
			printf("setsid error");
		if ((fds = ptys_open(fdm, pts_name)) < 0)
			printf("can't open slave pty");
		close(fdm);

#if defined(TIOCSCTTY) && !defined(CIBAUD)
	// 4.3+BSD way to acquire controlling terminal
   // !CIBAUD to avoid doing this under SunOS
	if (ioctl(fds, TIOCSCTTY, (char *) 0) < 0)
		printf("TIOCSCTTY error");
#endif

      if(slave_termios != NULL)
		{
			if (tcsetattr(fds, TCSANOW, slave_termios) < 0)
				printf("tcsetattr error on slave pty");
		}
		if (slave_winsize != NULL)
		{
			if (ioctl(fds, TIOCSWINSZ, slave_winsize) < 0)
				printf("TIOCSWINSZ error on slave pty");
		}
		if (dup2(fds, STDIN_FILENO) != STDIN_FILENO)
			printf("dup2 error to stdin");
		if (dup2(fds, STDOUT_FILENO) != STDOUT_FILENO)
			printf("dup2 error to stdout");
		if (dup2(fds, STDERR_FILENO) != STDERR_FILENO)
			printf("dup2 error to stderr");
		if (fds > STDERR_FILENO)
			close(fds);
		return(0);
	}
	else
	{
		*ptrfdm = fdm;
		return(pid);
	}
}

void loop(int ptym, int ignoreeof)
{
	pid_t child;
	int nread, num;
	char buff[BUFFSIZE];

	if ((child = fork()) < 0)
	{
		printf("fork error");
	}
	else if (child == 0)
	{
		for ( ; ; )
		{
			if ((nread = farm9crypt_read(STDIN_FILENO, buff, BUFFSIZE)) < 0)
				printf("read error from stdin");
			else if (nread == 0)
				break;
 
		// Detect certain strings to ensure we don't do any special processing
        if (strncmp(buff, "exit", 4) == 0)
			{
			  if (write(ptym, buff, nread) != nread)
				printf("write error to master pty");
			  return;
			}

		   if (write(ptym, buff, nread) != nread)
				printf("write error to master pty");
		}
		if (ignoreeof == 0)
			kill(getppid(), SIGTERM);
		exit(0);
	}
	if (signal_intr(SIGTERM, sig_term) == SIG_ERR)
		printf("signal_intr error for sigterm");
	for( ; ; )
	{
		if ((nread = read(ptym, buff, BUFFSIZE)) <= 0)
			break;
		if (farm9crypt_write(STDOUT_FILENO, buff, nread) != nread)
			printf("write error to stdout");
	}
	if (sigcaught == 0)
		kill(child, SIGTERM);
	return;
}

static void sig_term(int signo)
{
	sigcaught = 1;
	return;
}

static void set_noecho(int fd)
{
	struct termios stermios;
	
	if (tcgetattr(fd, &stermios) < 0)
		printf("tcgetattr error");
	stermios.c_lflag &= ~ (ECHO | ECHOE | ECHOK | ECHONL);
	stermios.c_oflag &= ~ (ONLCR);
	if (tcsetattr(fd, TCSANOW, &stermios) < 0)
		printf("tcsetattr error");
}

void doexec(int fdm, char *shell)
{
	int c, ignoreeof, interactive, noecho, verbose;
	pid_t pid;
	char slave_name[20];
	struct termios orig_termios;
	struct winsize size;

	interactive = isatty(STDIN_FILENO);
	ignoreeof = 1;
	noecho = 1;
	opterr = 0;

   dup2(fdm, 0);
	close(fdm);
	dup2(0,1);
	dup2(0,2);
	if (interactive)
	{
		if (tcgetattr(STDIN_FILENO, &orig_termios) < 0)
			printf("tcgetattr error on stdin");
		if (ioctl(STDIN_FILENO, TIOCGWINSZ, (char *) &size) < 0)
			printf("TIOCGWINSZ error");
		pid = pty_fork(&fdm, slave_name, &orig_termios, &size);
	}
	else
		pid = pty_fork(&fdm, slave_name, NULL, NULL);
	if (pid < 0)
		printf("fork error");
	else if (pid == 0)
	{
		if (noecho)
			set_noecho(STDIN_FILENO);
		putenv("TERM=dumb");
		if (execl(shell, "shell", NULL) < 0)
			printf("can't execute: %s", shell);
	}
	loop(fdm, ignoreeof);
	kill(getpid(), SIGTERM);
}
